import datetime
import os
import shutil

from metabolights_utils.utils.search_utils import get_isa_metadata_files


def create_audit_folder(
    src_root_path: str,
    target_root_path: str,
    folder_suffix: None | str = "BACKUP",
    folder_prefix: None | str = None,
    timestamp_format: str = "%Y-%m-%d_%H-%M-%S",
) -> None | str:
    metadata_files_list = get_isa_metadata_files(
        folder_path=src_root_path, recursive=False
    )
    metadata_files_list.sort()

    if metadata_files_list:
        base = datetime.datetime.now(datetime.timezone.utc).strftime(timestamp_format)
        folder_name = f"{base}_{folder_suffix}" if folder_suffix else base
        folder_name = f"{folder_prefix}_{folder_name}" if folder_prefix else folder_name

        target_folder_path = os.path.join(target_root_path, folder_name)

        os.makedirs(target_folder_path, exist_ok=True)

        for file in metadata_files_list:
            basename = os.path.basename(file)
            target_file = os.path.join(target_folder_path, basename)
            shutil.copy2(file, target_file, follow_symlinks=False)
        return target_folder_path
    return None


def copy_isa_metadata_files(
    src_folder_path: str, target_folder_path: str
) -> None | str:
    metadata_files_list = get_isa_metadata_files(
        folder_path=src_folder_path, recursive=False
    )
    metadata_files_list.sort()

    if metadata_files_list:
        os.makedirs(target_folder_path, exist_ok=True)
        for file in metadata_files_list:
            basename = os.path.basename(file)
            target_file = os.path.join(target_folder_path, basename)
            shutil.copy2(file, target_file, follow_symlinks=False)
        return target_folder_path
    return None



from typing import List

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.models.common import MetabolightsBaseModel
from metabolights_utils.models.parser.enums import ParserMessageType


class ParserMessage(MetabolightsBaseModel):
    type: Annotated[ParserMessageType, Field(description="")] = (
        ParserMessageType.WARNING
    )
    short: Annotated[str, Field(description="")] = ""
    detail: Annotated[str, Field(description="")] = ""
    section: Annotated[
        str, Field(description="Section name if it is investigation file")
    ] = ""
    line: Annotated[str, Field(description="")] = ""
    column: Annotated[str, Field(description="")] = (
        "Column name in ISA metadata table file"
    )

    def __str__(self) -> str:
        return "\t".join(
            [
                self.type,
                f"section: {self.section}",
                f"line: {self.line}",
                f"column: {self.column}",
                self.short,
                f"detail: {self.detail}",
            ]
        )


class ParserReport(MetabolightsBaseModel):
    messages: Annotated[List[ParserMessage], Field(description="")] = []



import pathlib
import uuid
from typing import List, Set

from metabolights_utils.tsv import model as actions
from metabolights_utils.tsv.actions.base import BaseTsvAction


class CopyRowTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvCopyRowAction,
    ) -> actions.TsvActionResult:
        result: actions.TsvActionResult = actions.TsvActionResult(action=action)
        if action.action_type != actions.TsvActionType.COPY_ROW:
            result.message = "Action name is not valid"
            return result

        action: actions.TsvCopyRowAction = action
        target_row_indices: List[int] = action.target_row_indices
        selected_column_indices: Set[int] = set(action.selected_column_indices)
        if not target_row_indices:
            result.message = "There is no target row index"
            return result

        source_index = action.source_row_index

        if not source_index or source_index < 0:
            result.message = "There is not row index"
            return result

        row_indices = target_row_indices.copy()
        row_indices.sort()

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value
        copied_row = None
        with open(source_file_path, "r", encoding="utf-8") as source:
            source.readline()
            row_index = 0
            for line in source:
                if row_index > source_index:
                    break
                if row_index == source_index:
                    copied_row = line.strip().split("\t")
                    break
                row_index += 1
        if not copied_row:
            result.message = "Row will be copied is not found"
            return result

        try:
            with open(source_file_path, "r", encoding="utf-8") as source:
                header_line = source.readline()

                with open(target_file_path, "w", encoding="utf-8") as target:
                    target.write(header_line)
                    row_index = 0
                    for line in source:
                        if row_index in row_indices:
                            if not selected_column_indices:
                                self.write_row(target, copied_row)
                            else:
                                new_row = line.strip().split("\t")
                                for index in selected_column_indices:
                                    new_row[index] = copied_row[index]
                                    self.write_row(target, new_row)
                            row_indices.remove(row_index)
                        else:
                            target.write(line)
                        row_index += 1

                    if len(row_indices):
                        result.message = (
                            f"Invalid target row indices {', '.join(row_indices)}"
                        )
                        return result
            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result



import re

import unidecode


def sanitise_filename(
    filename: str,
    replacement_chars_pattern: None | str = None,
    allow_spaces: bool = False,
) -> str:
    if not filename or not filename.strip():
        return ""

    if not replacement_chars_pattern:
        replacement_chars_pattern = "[^/a-zA-Z0-9_.-]"
    filename = unidecode(filename.strip())
    filename = filename.replace("+", "_PLUS_")
    if not allow_spaces:
        filename = "".join([x.strip() for x in filename.strip().split() if x.strip()])

    filename = re.sub(replacement_chars_pattern, "_", filename)
    return filename



import pathlib
from abc import ABC
from io import IOBase
from typing import Tuple, Union


class BaseIsaFile(ABC):
    def _get_file_path(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
    ) -> Tuple[Union[str, pathlib.Path, IOBase], str]:
        if not file_buffer_or_path:
            return None
        if isinstance(file_buffer_or_path, IOBase):
            return file_buffer_or_path, file_buffer_or_path.name
        if isinstance(file_buffer_or_path, pathlib.Path):
            return file_buffer_or_path, str(file_buffer_or_path)
        elif isinstance(file_buffer_or_path, str):
            return file_buffer_or_path, file_buffer_or_path

    def _get_file_buffer(
        self, file: Union[str, pathlib.Path, IOBase], encoding="utf-8"
    ) -> IOBase:
        if isinstance(file, IOBase):
            file_buffer = file
        elif isinstance(file, pathlib.Path):
            file_buffer = file.open(mode="r", encoding=encoding)
        elif isinstance(file, str):
            file_buffer = open(file=file, mode="r", encoding=encoding)
        else:
            raise ValueError("file type is not defined")

        return file_buffer

    def _close_file(
        self,
        file_buffer_or_path: Union[str, pathlib.Path, IOBase],
    ):
        # If we opened file, close it
        if (
            not file_buffer_or_path
            and not file_buffer_or_path
            and (
                isinstance(file_buffer_or_path, str)
                or isinstance(file_buffer_or_path, pathlib.Path)
            )
        ):
            file_buffer_or_path.close()



from typing_extensions import Annotated

from pydantic import Field

from metabolights_utils.common import CamelCaseModel
from metabolights_utils.models.enums import GenericMessageType


class MetabolightsBaseModel(CamelCaseModel):
    pass


class GenericMessage(MetabolightsBaseModel):
    type: Annotated[
        GenericMessageType, Field(description="")
    ] = GenericMessageType.ERROR
    short: Annotated[str, Field(description="")] = ""
    detail: Annotated[str, Field(description="")] = ""

    def __str__(self) -> str:
        value = f"{self.type}\tshort: {self.short}\tdetail: {self.detail}"
        return value



from metabolights_utils import common as metabolights_common
from metabolights_utils.isatab import Reader, Writer, reader, writer
from metabolights_utils.isatab.default import assay_file as default_assay_file_reader
from metabolights_utils.isatab.default import (
    assignment_file as default_assignment_file_reader,
)
from metabolights_utils.isatab.default import base_isa_file, base_isa_table_file
from metabolights_utils.isatab.default import factory as default_factory
from metabolights_utils.isatab.default import (
    investigation_file as default_investigation_file_operations,
)
from metabolights_utils.isatab.default import sample_file as default_sample_file_reader
from metabolights_utils.isatab.default import writer as default_isa_writer
from metabolights_utils.isatab.default.parser import common as default_parser_commons
from metabolights_utils.isatab.default.parser import (
    investigation_parser,
    isa_table_parser,
)
from metabolights_utils.models import common as models_common
from metabolights_utils.models import enums as models_enum
from metabolights_utils.models.isa import assay_file, assignment_file
from metabolights_utils.models.isa import common as isa_commons
from metabolights_utils.models.isa import enums as isa_enums
from metabolights_utils.models.isa import investigation_file, samples_file
from metabolights_utils.models.metabolights import metabolights_study
from metabolights_utils.models.parser import common as parser_common
from metabolights_utils.models.parser import enums as parser_enums
from metabolights_utils.tsv import filter as tsv_filter
from metabolights_utils.tsv import model as tsv_model
from metabolights_utils.tsv import sort as tsv_sort
from metabolights_utils.tsv import tsv_file_updater
from metabolights_utils.tsv import utils as tsv_utils
from metabolights_utils.tsv.actions import add_column, add_row
from metabolights_utils.tsv.actions import base as tsv_actions_base
from metabolights_utils.tsv.actions import (
    copy_column,
    copy_row,
    delete_column,
    delete_row,
    move_column,
    move_row,
    update_cell,
    update_column,
    update_column_header,
    update_row,
)
from metabolights_utils.utils import (
    audit_utils,
    filename_utils,
    hash_utils,
    search_utils,
)
from metabolights_utils.utils.audit_utils import create_audit_folder
from metabolights_utils.utils.hash_utils import (
    get_isa_metadata_folder_hash,
    sha256sum,
    IsaMetadataFolderHash,
    EMPTY_FILE_HASH,
)

__all__ = [
    "Reader",
    "Writer",
    "metabolights_common",
    "metabolights_study",
    "parser_common",
    "parser_enums",
    "models_common",
    "models_enum",
    "assay_file",
    "assignment_file",
    "isa_commons",
    "isa_enums",
    "investigation_file",
    "samples_file",
    "reader",
    "writer",
    "default_assay_file_reader",
    "default_assignment_file_reader",
    "base_isa_file",
    "base_isa_table_file",
    "default_factory",
    "default_investigation_file_operations",
    "default_sample_file_reader",
    "default_isa_writer",
    "default_parser_commons",
    "investigation_parser",
    "isa_table_parser",
    # TSV
    "tsv_filter",
    "tsv_model",
    "tsv_sort",
    "tsv_file_updater",
    "tsv_utils",
    "add_column",
    "add_row",
    "tsv_actions_base",
    "copy_column",
    "copy_row",
    "delete_column",
    "delete_row",
    "move_column",
    "move_row",
    "update_cell",
    "update_column_header",
    "update_column",
    "update_row",
    # UTILS
    "audit_utils",
    "hash_utils",
    "filename_utils",
    "search_utils",
    # audit utils
    "create_audit_folder"
    # hash utils
    "get_isa_metadata_folder_hash",
    "sha256sum",
    "IsaMetadataFolderHash",
    "EMPTY_FILE_HASH",
]



import glob
import hashlib
import os
from typing import Dict

from pydantic import BaseModel

from metabolights_utils.utils.search_utils import get_isa_metadata_files

EMPTY_FILE_HASH = hashlib.sha256("".encode()).hexdigest()


class IsaMetadataFolderHash(BaseModel):
    folder_sha256: str = ""
    files_sha256: Dict[str, str] = {}


def sha256sum(filepath: str):
    if not filepath or not os.path.exists(filepath):
        return EMPTY_FILE_HASH

    sha256_hash = hashlib.sha256()
    with open(filepath, "rb") as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def get_isa_metadata_folder_hash(folder_path: str) -> IsaMetadataFolderHash:
    files: list[str] = get_isa_metadata_files(folder_path, recursive=False)
    files.sort()
    result = IsaMetadataFolderHash()
    hashes = []
    for file in files:
        basename = os.path.basename(file)
        hash = sha256sum(file)
        result.files_sha256[basename] = hash
        hashes.append(f"{basename}:{hash}")
    hash_bytes = ",".join(hashes).encode("utf-8")
    result.folder_sha256 = hashlib.sha256(hash_bytes).hexdigest()
    return result



import hashlib


def calculate_sha256(file: str) -> str:
    sha256_hash = hashlib.sha256()
    with open(file, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            sha256_hash.update(chunk)
    return sha256_hash.hexdigest()



import pathlib
import uuid
from typing import Dict, List

from metabolights_utils.tsv import model as actions
from metabolights_utils.tsv.actions.base import BaseTsvAction


class UpdateCellsTsvAction(BaseTsvAction):
    def apply_action(
        self,
        source_file_path: pathlib.Path,
        target_file_path: pathlib.Path,
        action: actions.TsvUpdateCellsAction,
    ) -> actions.TsvActionResult:
        result: actions.TsvActionResult = actions.TsvActionResult(action=action)
        if action.action_type != actions.TsvActionType.UPDATE_CELL_DATA:
            result.message = "Action name is not valid"
            return result

        action: actions.TsvUpdateCellsAction = action

        cells: List[actions.TsvCellData] = action.cells if action.cells else []

        if not cells:
            result.message = "There is no cell"
            return result

        cells.sort(key=lambda x: (x.row_index, x.column_index))
        row_data: Dict[idx, actions.TsvRowData] = {}
        for idx, val in enumerate(cells):
            row_idx = val.row_index
            col_idx = val.column_index
            if row_idx not in row_data:
                row_data[row_idx] = actions.TsvRowData()
            data: actions.TsvRowData = row_data[row_idx]
            data.values[col_idx] = val.value

        row_indices = {x.row_index for x in cells}
        column_indices = {x.column_index for x in cells}

        if not action.id:
            uuid_value = str(uuid.uuid4().hex)
            action.id = uuid_value

        try:
            with open(source_file_path, "r", encoding="utf-8") as source:
                header_line = source.readline()
                header_names = header_line.strip().split("\t")
                invalid_column_indices = [
                    x for x in column_indices if x < 0 or x > len(header_names)
                ]
                if invalid_column_indices:
                    result.message = (
                        f"Invalid column indices: {', '.join(invalid_column_indices)}"
                    )
                    return result

                with open(target_file_path, "w", encoding="utf-8") as target:
                    self.write_row(target, header_names)
                    row_index = 0
                    for line in source:
                        row = line.strip().split("\t")
                        if row_index in row_data:
                            row_data_item = row_data[row_index]
                            for column_idx in row_data_item.values:
                                row[column_idx] = row_data_item.values[column_idx]
                            if row_index in row_indices:
                                row_indices.remove(row_index)

                        self.write_row(target, row)
                        row_index += 1
                    if row_indices:
                        result.message = (
                            f"Invalid row indices: {', '.join(row_indices)}"
                        )
                        return result
            result.success = True
        except Exception as exc:
            result.message = f"{str(exc)}"

        return result



import re
import sys
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum, EnumMeta
from typing import Any, Dict, List, Set, Union

from pydantic import Field
from typing_extensions import Annotated

from metabolights_utils.common import CamelCaseModel


class TsvFilterException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class FilterDataType(str, Enum):
    AUTO = "AUTO"
    STRING = "STRING"
    INTEGER = "INTEGER"
    FLOAT = "FLOAT"
    DATETIME = "DATETIME"


class FilterOperation(str, Enum):
    CONTAINS = "like"
    EQUAL = "eq"
    STARTSWITH = "startswith"
    ENDSWITH = "endswith"
    GREATER = "gt"
    GREATER_EQUAL = "ge"
    LESS = "lt"
    LESS_EQUAL = "le"
    REGEX = "regex"
    EMPTY = "empty"
    CUSTOM = "custom"


class TsvFileFilterOption(CamelCaseModel):
    search_columns: Annotated[
        List[str],
        Field(
            description="Search column names. Search in any column if it is None or empty",
        ),
    ] = []

    search_ignore_columns: Annotated[
        List[str],
        Field(
            description="Ignore selected columns if search column is None or empty",
        ),
    ] = []

    operation: Annotated[
        FilterOperation, Field(description="Filter operation")
    ] = FilterOperation.CONTAINS

    parameter: Annotated[
        Union[str, int, float, datetime],
        Field(description="Parameter for the filter operation"),
    ] = ""

    data_type: Annotated[
        FilterDataType, Field(description="Apply filter with the given data type")
    ] = FilterDataType.AUTO

    case_sensitive: Annotated[
        bool,
        Field(description="String operations will be applied in case sensitive mode"),
    ] = True

    negate_result: Annotated[
        bool, Field(description="Reverse result after apply operation")
    ] = False

    default_datetime_pattern: Annotated[
        str, Field(description="Convert string value to datetime using this pattern")
    ] = "%d/%m/%Y"

    custom_filter_arguments: Annotated[
        Dict[str, Any], Field(description="Additional arguments for custom filters")
    ] = {}

    custom_filter_name: Annotated[
        str, Field(description="Registered custom filter name")
    ] = ""


class Filter(ABC):
    def __init__(
        self,
        filter_option: TsvFileFilterOption,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
    ) -> None:
        self.column_name_indices = column_name_indices
        self.column_indices = column_indices
        self.filter_option: TsvFileFilterOption = filter_option

        if not filter_option:
            raise TsvFilterException("Filter option is not valid")

        if self.filter_option.search_columns:
            invalid_columns = []
            for column in self.filter_option.search_columns:
                if not column in self.column_name_indices:
                    invalid_columns.append(column)
            if invalid_columns:
                raise TsvFilterException(
                    f"Invalid search columns: {', '.join(invalid_columns)}"
                )
        else:
            self.filter_option.search_columns = []

        if self.filter_option.search_ignore_columns:
            invalid_columns = []
            for column in self.filter_option.search_ignore_columns:
                if not column in self.column_name_indices:
                    invalid_columns.append(column)
            if invalid_columns:
                raise TsvFilterException(
                    f"Invalid search ignore columns: {', '.join(invalid_columns)}"
                )
        else:
            self.filter_option.search_ignore_columns = []

        self.target_column_indices: Set[int] = set(
            self.column_name_indices[column_name]
            for column_name in self.column_name_indices
            if column_name in self.filter_option.search_columns
        )
        if not self.filter_option.search_columns:
            all_indices: Set[int] = set(self.column_indices.keys())
            ignore_column_indices: Set[int] = set(
                self.column_name_indices[column_name]
                for column_name in self.column_name_indices
                if column_name in self.filter_option.search_ignore_columns
            )
            self.target_column_indices = all_indices.difference(ignore_column_indices)

        self.parameter: Union[int, str, float, datetime] = (
            filter_option.parameter if filter_option.parameter is not None else ""
        )
        if not filter_option.case_sensitive and isinstance(self.parameter, str):
            self.parameter = filter_option.parameter.lower()

        self.parameter_data_type = self.filter_option.data_type

        if filter_option.operation != FilterOperation.CUSTOM:
            if self.filter_option.data_type == FilterDataType.AUTO:
                if isinstance(self.filter_option.parameter, float):
                    self.parameter_data_type = FilterDataType.FLOAT
                elif isinstance(self.filter_option.parameter, int):
                    self.parameter_data_type = FilterDataType.INTEGER
                elif isinstance(self.filter_option.parameter, datetime):
                    self.parameter_data_type = FilterDataType.DATETIME
                else:
                    self.parameter_data_type = FilterDataType.STRING
            self.parameter = self.convert_to_selected_data_type(self.parameter)

    def convert_to_selected_data_type(self, value):
        if self.parameter_data_type == FilterDataType.FLOAT:
            return float(value)
        elif self.parameter_data_type == FilterDataType.INTEGER:
            return int(value)
        elif self.parameter_data_type == FilterDataType.DATETIME:
            pattern = self.filter_option.default_datetime_pattern
            return datetime.strptime(value, pattern)
        if not value:
            return ""

        return value if self.filter_option.case_sensitive else value.lower()

    def get_updated_value(self, value: str):
        if value is None:
            return ""
        return value if self.filter_option.case_sensitive else value.lower()

    def filter(self, row: List[str]) -> bool:
        match = False
        for index in self.target_column_indices:
            if self.evaluate(row[index]):
                match = True
                break

        return not match if self.filter_option.negate_result else match

    @abstractmethod
    def evaluate(self, row_value: str):
        pass


class CustomFilter(Filter, ABC):
    def __init__(
        self,
        filter_option: TsvFileFilterOption,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
        **kwargs,
    ) -> None:
        super().__init__(filter_option, column_name_indices, column_indices)
        self.kwargs = kwargs


class FilterRegistry:
    custom_filters: Dict[str, Any] = {}
    default_filters: Dict[str, Any] = {}

    @classmethod
    def register_filter(cls, name: str, filter_class: Filter) -> None:
        cls.default_filters[name] = filter_class

    @classmethod
    def get_filter(
        cls,
        filter_option: TsvFileFilterOption,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
    ) -> Filter:
        name = ""
        if filter_option.operation == FilterOperation.CUSTOM:
            name = filter_option.custom_filter_name
            if name in cls.custom_filters:
                return cls.custom_filters[name](
                    filter_option, column_name_indices, column_indices
                )
        else:
            name = filter_option.operation.value
            if name in cls.default_filters:
                return cls.default_filters[name](
                    filter_option, column_name_indices, column_indices
                )
        raise TsvFilterException(f"Filter name {name} is not registered")

    @classmethod
    def register_custom_filter(cls, name: str, filter_class: CustomFilter) -> None:
        cls.custom_filters[name] = filter_class

    @classmethod
    def unregister_custom_filter(cls, name: str) -> None:
        if name in cls.custom_filters:
            del cls.custom_filters[name]

    # @classmethod
    # def get_custom_filter(
    #     cls,
    #     filter_option: TsvFileFilterOption,
    #     column_name_indices: Dict[str, int],
    #     column_indices: Dict[int, str],
    # ) -> Filter:
    #     name = filter_option.custom_filter_name
    #     if name and name in cls.custom_filters:
    #         return cls.custom_filters[name](
    #             filter_option, column_name_indices, column_indices
    #         )
    #     return None


# FILTERS


# class CustomFilterProxy(Filter):
#     def __init__(
#         self,
#         filter_option: TsvFileFilterOption,
#         column_name_indices: Dict[str, int],
#         column_indices: Dict[int, str],
#     ) -> None:
#         super().__init__(filter_option, column_name_indices, column_indices)
#         self.custom_filter: CustomFilter = FilterRegistry.get_custom_filter(
#             self.filter_option, column_name_indices, column_indices
#         )

#     def evaluate(self, row_value: str) -> bool:
#         if not self.custom_filter:
#             return False
#         return self.custom_filter.evaluate(row_value)


class ContainsFilter(Filter):
    def evaluate(self, row_value: str) -> bool:
        value = self.get_updated_value(row_value)
        return True if self.parameter in value else False


class StartsWithFilter(Filter):
    def evaluate(self, row_value: str) -> bool:
        value = self.get_updated_value(row_value)
        return True if value.startswith(self.parameter) else False


class EndsWithFilter(Filter):
    def evaluate(self, row_value: str) -> bool:
        value = self.get_updated_value(row_value)
        return True if value.endswith(self.parameter) else False


class EqualFilter(Filter):
    def evaluate(self, row_value: str) -> bool:
        value = self.get_updated_value(row_value)
        return True if value == self.parameter else False


class GreaterFilter(Filter):
    def evaluate(self, row_value: str) -> bool:
        if row_value:
            try:
                value = self.convert_to_selected_data_type(row_value)
                return True if value > self.parameter else False
            except Exception:
                return False
        return False


class GreaterEqualFilter(Filter):
    def evaluate(self, row_value: str) -> bool:
        if row_value:
            try:
                value = self.convert_to_selected_data_type(row_value)
                return True if value >= self.parameter else False
            except Exception:
                return False
        return False


class LessFilter(Filter):
    def evaluate(self, row_value: str) -> bool:
        if row_value:
            try:
                value = self.convert_to_selected_data_type(row_value)
                return True if value < self.parameter else False
            except Exception:
                return False
        return False


class LessEqualFilter(Filter):
    def evaluate(self, row_value: str) -> bool:
        if row_value:
            try:
                value = self.convert_to_selected_data_type(row_value)
                return True if value <= self.parameter else False
            except Exception:
                return False
        return False


class EmptyFilter(Filter):
    def evaluate(self, row_value: str) -> bool:
        return True if not row_value else False


class RegexFilter(Filter):
    def evaluate(self, row_value: str) -> bool:
        value = self.get_updated_value(row_value)
        return True if re.match(self.parameter, value) else False


FilterRegistry.register_filter(FilterOperation.CONTAINS.value, ContainsFilter)
FilterRegistry.register_filter(FilterOperation.STARTSWITH.value, StartsWithFilter)
FilterRegistry.register_filter(FilterOperation.ENDSWITH.value, EndsWithFilter)
FilterRegistry.register_filter(FilterOperation.EQUAL.value, EqualFilter)
FilterRegistry.register_filter(FilterOperation.GREATER.value, GreaterFilter)
FilterRegistry.register_filter(FilterOperation.GREATER_EQUAL.value, GreaterEqualFilter)
FilterRegistry.register_filter(FilterOperation.LESS.value, LessFilter)
FilterRegistry.register_filter(FilterOperation.LESS_EQUAL.value, LessEqualFilter)
FilterRegistry.register_filter(FilterOperation.EMPTY.value, EmptyFilter)
FilterRegistry.register_filter(FilterOperation.REGEX.value, RegexFilter)


class ValidNumberCustomFilter(CustomFilter):
    def evaluate(self, row_value: str) -> bool:
        try:
            _ = float(row_value)
            return True
        except Exception:
            return False


class ValidDatetimeCustomFilter(CustomFilter):
    def evaluate(self, row_value: str) -> bool:
        try:
            value = datetime.strptime(
                row_value, self.filter_option.default_datetime_pattern
            )
            return True if value else False
        except Exception:
            return False


class EnumContainsCustomFilter(CustomFilter):
    def __init__(
        self,
        filter_option: TsvFileFilterOption,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
    ) -> None:
        super().__init__(filter_option, column_name_indices, column_indices)

        if "enum-class" in filter_option.custom_filter_arguments:
            self.enum_class = filter_option.custom_filter_arguments["enum-class"]
            try:
                if isinstance(self.enum_class, str):
                    splitted_value = self.enum_class.split(":")

                    if len(splitted_value) == 2:
                        self.enum_class = getattr(
                            sys.modules[splitted_value[0]], splitted_value[1]
                        )
                    else:
                        raise TsvFilterException(
                            "Selected enum class name is not valid. Use format <module name>:<enum class name>"
                        )
            except Exception as exc:
                if isinstance(exc, TsvFilterException):
                    raise exc
                raise TsvFilterException(
                    "Selected class name is not converted to enum"
                ) from exc
        if not isinstance(self.enum_class, EnumMeta):
            raise TsvFilterException("selected class is not enum")

        self.enum_string_map: Dict[str, str] = {}
        if "enum-value-map" in filter_option.custom_filter_arguments:
            self.enum_value_map = filter_option.custom_filter_arguments[
                "enum-value-map"
            ]
            if not isinstance(self.enum_value_map, dict) or not self.enum_value_map:
                raise TsvFilterException("Enum map is not valid dict")

            for key in self.enum_value_map:
                value = self.enum_value_map[key]
                if self.filter_option.case_sensitive:
                    self.enum_string_map[str(key)] = str(value)
                else:
                    self.enum_string_map[str(key)] = str(value).lower()

            enum_values = {str(e.value) for e in self.enum_class}
            missing_values = [x for x in enum_values if x not in self.enum_string_map]
            if missing_values:
                raise TsvFilterException(
                    f"There are some missing values to map an enum {', '.join(missing_values)}"
                )

    def evaluate(self, row_value: str) -> bool:
        try:
            if self.filter_option.case_sensitive:
                value = self.parameter in self.enum_string_map[row_value]
            else:
                value = self.parameter in self.enum_string_map[str(row_value).lower()]
            return True if value else False
        except Exception:
            return False


class BetweenEqualCustomFilter(CustomFilter):
    def __init__(
        self,
        filter_option: TsvFileFilterOption,
        column_name_indices: Dict[str, int],
        column_indices: Dict[int, str],
    ) -> None:
        super().__init__(filter_option, column_name_indices, column_indices)
        self.min = None
        self.max = None
        if "min" in self.filter_option.custom_filter_arguments:
            self.min = self.filter_option.custom_filter_arguments["min"]
        if "max" in self.filter_option.custom_filter_arguments:
            self.max = self.filter_option.custom_filter_arguments["max"]
        self.min = self.convert_to_selected_data_type(self.min)
        self.max = self.convert_to_selected_data_type(self.max)
        if not self.min or not self.max:
            raise TsvFilterException("min and max values are not valid.")

        if self.filter_option.data_type == FilterDataType.AUTO:
            if isinstance(self.min, float):
                self.parameter_data_type = FilterDataType.FLOAT
            elif isinstance(self.min, int):
                self.parameter_data_type = FilterDataType.INTEGER
            elif isinstance(self.min, datetime):
                self.parameter_data_type = FilterDataType.DATETIME
            else:
                self.parameter_data_type = FilterDataType.STRING

    def evaluate(self, row_value: str) -> bool:
        if row_value:
            try:
                value = self.convert_to_selected_data_type(row_value)
                return True if self.min <= value <= self.max else False
            except Exception:
                return False
        return False


FilterRegistry.register_custom_filter("between-equal", BetweenEqualCustomFilter)
FilterRegistry.register_custom_filter("valid-datetime", ValidDatetimeCustomFilter)
FilterRegistry.register_custom_filter("valid-number", ValidNumberCustomFilter)
FilterRegistry.register_custom_filter("enum-contains", EnumContainsCustomFilter)
